from collections import Counter

import torch
import numpy as np
import torchvision
import matplotlib
import matplotlib.pyplot as plt
from torch.hub import download_url_to_file
import os
import pickle
import torch.utils.data
import torch.nn.functional as F
from torch.utils.data import Subset
from tqdm import tqdm
import sklearn.model_selection
import torch.utils.data

matplotlib.use('TkAgg')
plt.rcParams["figure.figsize"] = (15, 5)
plt.style.use('dark_background')

LEARNING_RATE = 1e-4
BATCH_SIZE = 128
MAX_LEN = 200
TRAIN_TEST_SPLIT = 0.7
DEVICE = 'cpu'

if torch.cuda.is_available():
    DEVICE = 'cuda'
    MAX_LEN = 0

class Dataset(torch.utils.data.Dataset):
    def __init__(self):
        super().__init__()
        path_dataset = '../data/Fruits28.pkl'
        if not os.path.exists(path_dataset):
            pass
            os.makedirs('../data', exist_ok=True)
            download_url_to_file(
                'http://share.yellowrobot.xyz/1645110979-deep-learning-intro-2022-q1/Fruits28.pkl',
                path_dataset,
                progress=True
            )
        with open(path_dataset, 'rb') as fp:
            X, Y, self.labels = pickle.load(fp)
        self.Y_idx = Y

        Y_counter = Counter(Y)
        Y_counts = np.array(list(Y_counter.values()))
        self.Y_weights = (1.0 / Y_counts) * np.sum(Y_counts)

        X = torch.from_numpy(np.array(X).astype(np.float32))
        self.X = X.permute(0, 3, 1, 2)
        self.input_size = self.X.size(-1)
        Y = torch.LongTensor(Y)
        self.Y = F.one_hot(Y)

    def __len__(self):
        return len(self.X)

    def __getitem__(self, idx):
        x = self.X[idx]
        y = self.Y[idx]

        return x, y


dataset_full = Dataset()
train_test_split = int(len(dataset_full) * TRAIN_TEST_SPLIT)

idxes_train, idxes_test = sklearn.model_selection.train_test_split(
    np.arange(len(dataset_full)),
    train_size=train_test_split,
    test_size=len(dataset_full) - train_test_split,
    stratify=dataset_full.Y_idx,
    random_state=0
)

# For debugging
if MAX_LEN:
    idxes_train = idxes_train[:MAX_LEN]
    idxes_test = idxes_test[:MAX_LEN]

dataset_train = Subset(dataset_full, idxes_train)
dataset_test = Subset(dataset_full, idxes_test)

dataloader_train = torch.utils.data.DataLoader(
    dataset=dataset_train,
    batch_size=BATCH_SIZE,
    shuffle=True,
    drop_last=(len(dataset_train) % BATCH_SIZE == 1)
)

dataloader_test = torch.utils.data.DataLoader(
    dataset=dataset_test,
    batch_size=BATCH_SIZE,
    shuffle=False,
    drop_last=(len(dataset_test) % BATCH_SIZE == 1)
)


class DenseBlock(torch.nn.Module):
    def __init__(self):
        super().__init__()

        self.bn1 = torch.nn.BatchNorm2d(num_features=32)
        self.bn2 = torch.nn.BatchNorm2d(num_features=64)
        self.bn3 = torch.nn.BatchNorm2d(num_features=96)
        self.bn4 = torch.nn.BatchNorm2d(num_features=128)

        # out channels fixed by layer count and each layer out channels, 4 * 32 = 128
        self.conv1 = torch.nn.Conv2d(in_channels=32, out_channels=32, kernel_size=3, stride=1, padding=1)
        self.conv2 = torch.nn.Conv2d(in_channels=64, out_channels=32, kernel_size=3, stride=1, padding=1)
        self.conv3 = torch.nn.Conv2d(in_channels=96, out_channels=32, kernel_size=3, stride=1, padding=1)
        self.conv4 = torch.nn.Conv2d(in_channels=128, out_channels=32, kernel_size=3, stride=1, padding=1)

    def forward(self, x):

        conv1 = self.conv1(self.bn1(F.relu(x)))
        conv2_in = torch.cat([x, conv1], dim=1)

        conv2 = self.conv2(self.bn2(F.relu(conv2_in)))
        conv3_in = torch.cat([x, conv1, conv2], dim=1)

        conv3 = self.conv3(self.bn3(F.relu(conv3_in)))
        conv4_in = torch.cat([x, conv1, conv2, conv3], dim=1)

        conv4 = self.conv4(self.bn4(F.relu(conv4_in)))
        output = torch.cat([x, conv1, conv2, conv3, conv4], dim=1)

        return output


class TransitionLayer(torch.nn.Module):
    def __init__(self, in_channels, out_channels):
        super().__init__()

        self.bn = torch.nn.BatchNorm2d(num_features=out_channels)
        self.conv = torch.nn.Conv2d(in_channels=in_channels, out_channels=out_channels, kernel_size=1, bias=False)
        self.avg_pool = torch.nn.AvgPool2d(kernel_size=2, stride=2, padding=0)

    def forward(self, x):
        bn = self.bn(F.relu(self.conv(x)))
        out = self.avg_pool(bn)

        return out


class DenseNet(torch.nn.Module):
    def __init__(self):
        super().__init__()

        self.conv1 = torch.nn.Conv2d(in_channels=3, out_channels=32, kernel_size=3, stride=1, padding=1, bias=False)

        self.denseblock1 = DenseBlock()
        self.denseblock2 = DenseBlock()
        self.denseblock3 = DenseBlock()

        self.transitionLayer1 = TransitionLayer(in_channels=160, out_channels=32)
        self.transitionLayer2 = TransitionLayer(in_channels=160, out_channels=32)
        self.transitionLayer3 = TransitionLayer(in_channels=160, out_channels=32)

        # Classifier
        self.linear = torch.nn.Linear(in_features=288, out_features=len(dataset_full.labels))

    def forward(self, x):
        out = self.conv1.forward(x)

        out = self.denseblock1.forward(out)
        out = self.transitionLayer1.forward(out)

        out = self.denseblock2.forward(out)
        out = self.transitionLayer2.forward(out)

        out = self.denseblock3.forward(out)
        out = self.transitionLayer3.forward(out)

        out = out.view(-1, 32*3*3)
        out = self.linear(out)
        out = F.softmax(out, dim=1)
        return out


# in deeper models using bias makes no difference and only increases execution time
class ResBlock(torch.nn.Module):
    def __init__(self, in_channels, out_channels, stride=1):
        super().__init__()
        self.conv1 = torch.nn.Conv2d(in_channels, out_channels, kernel_size=3, stride=stride, padding=1, bias=False)
        self.bn1 = torch.nn.BatchNorm2d(out_channels)
        self.conv2 = torch.nn.Conv2d(out_channels, out_channels, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn2 = torch.nn.BatchNorm2d(out_channels)

        self.is_bottleneck = False
        if stride != 1 or in_channels != out_channels:
            self.is_bottleneck = True
            self.shortcut = torch.nn.Conv2d(in_channels, out_channels, kernel_size=1, stride=stride, bias=False)

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = F.relu(out)
        out = self.bn1(out)

        out = self.conv2(out)

        if self.is_bottleneck:
            residual = self.shortcut(x)
        out += residual

        out = F.relu(out)
        out = self.bn2(out)
        return out


class ModelResNet(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.in_planes = 8
        self.block_size = 2

        # first convolution layer can have bigger kernel size
        self.conv1 = torch.nn.Conv2d(in_channels=3, out_channels=self.in_planes, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn1 = torch.nn.BatchNorm2d(self.in_planes)

        # make resblocks
        # common use, after bottleneck block at least one identity block
        self.identity_block_1 = ResBlock(in_channels=8, out_channels=8)
        self.identity_block_2 = ResBlock(in_channels=8, out_channels=8)

        self.bottleneck_block_1 = ResBlock(in_channels=8, out_channels=16, stride=2)
        self.identity_block_3 = ResBlock(in_channels=16, out_channels=16)

        self.bottleneck_block_2 = ResBlock(in_channels=16, out_channels=32, stride=2)
        self.identity_block_4 = ResBlock(in_channels=32, out_channels=32)

        self.bottleneck_block_3 = ResBlock(in_channels=32, out_channels=64, stride=2)
        self.identity_block_5 = ResBlock(in_channels=64, out_channels=64)

        self.linear = torch.nn.Linear(in_features=64, out_features=len(dataset_full.labels))

    def forward(self,x):
        out = self.bn1(F.relu(self.conv1(x)))

        out = self.identity_block_1(out)
        out = self.identity_block_2(out)
        out = self.bottleneck_block_1(out)
        out = self.identity_block_3(out)
        out = self.bottleneck_block_2(out)
        out = self.identity_block_4(out)
        out = self.bottleneck_block_3(out)
        out = self.identity_block_5(out)

        out = F.avg_pool2d(out, 4)
        out = out.view(out.size(0), -1)
        out = self.linear(out)
        out = F.softmax(out, dim=1)
        return out


class ModelReusableResNet(torch.nn.Module):
    def __init__(self, first_conv_channels=2, input_size=28, expansion=2, block_count=3, block_size=2):
        super().__init__()
        self.in_planes = first_conv_channels
        self.expansion = expansion
        self.block_count = block_count
        self.block_size = block_size

        self.output_size = input_size

        # first convolution layer can have bigger kernel size
        self.conv1 = torch.nn.Conv2d(in_channels=3, out_channels=self.in_planes, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn1 = torch.nn.BatchNorm2d(self.in_planes)

        self.blocks = torch.nn.ModuleList()
        for idx in range(self.block_count):
            self.blocks.append(
                self._make_layer(self.in_planes*self.expansion, self.block_size, stride=2)
            )

        flat_size = self.in_planes * self.output_size ** 2
        self.linear = torch.nn.Linear(flat_size, out_features=len(dataset_full.labels))

    def _make_layer(self, planes, num_blocks, stride):
        strides = [stride] + [1]*(num_blocks-1)
        layers = []
        for stride in strides:
            layers.append(ResBlock(self.in_planes, planes, stride))
            self.in_planes = planes
            self.output_size = int((self.output_size+2*1-(3-1)-1)/stride + 1)
        return torch.nn.Sequential(*layers)

    def forward(self,x):
        out = self.bn1(F.relu(self.conv1(x)))
        for block in self.blocks:
            out = block(out)
        out = out.view(out.size(0), -1)
        out = self.linear(out)
        out = F.softmax(out, dim=1)
        return out


class ModelConv(torch.nn.Module):
    def __init__(self):
        super().__init__()
        in_size = 28
        out_size = int((in_size - 5 + (2*1)) / 2 + 1)
        out_size = int((out_size - 2 + (2*0)) / 2 + 1)
        out_size = int((out_size - 3 + (2*1)) / 2 + 1)
        self.out_size = out_size
        self.encoder = torch.nn.Sequential(
            torch.nn.Conv2d(in_channels=1, out_channels=5, kernel_size=5, stride=2, padding=1),
            torch.nn.LeakyReLU(),
            torch.nn.BatchNorm2d(num_features=5),
            torch.nn.MaxPool2d(kernel_size=2, stride=2, padding=0),
            torch.nn.LeakyReLU(),
            torch.nn.BatchNorm2d(num_features=5),
            torch.nn.Conv2d(in_channels=5, out_channels=len(dataset_full.labels), kernel_size=3, stride=2, padding=1),
            torch.nn.ReLU(),
        )

    def forward(self, x):
        out = self.encoder.forward(x)
        out = F.adaptive_avg_pool2d(out, output_size=(1, 1))
        out = out.view(x.size(0), -1)
        out = torch.nn.functional.softmax(out, dim=1)
        return out

def print_model_size(model):
    total_param_size = 0
    for name, param in model.named_parameters():
        each_param_size = np.prod(param.size())
        total_param_size += each_param_size
    print(f'model size is {total_param_size}')

model = ModelReusableResNet()
print_model_size(model)

model = model.to(DEVICE)
optimizer = torch.optim.Adam(model.parameters(), lr=LEARNING_RATE)

metrics = {}
for stage in ['train', 'test']:
    for metric in [
        'loss',
        'acc'
    ]:
        metrics[f'{stage}_{metric}'] = []

for epoch in range(1, 100):
    for data_loader in [dataloader_train, dataloader_test]:
        metrics_epoch = {key: [] for key in metrics.keys()}

        stage = 'train'
        if data_loader == dataloader_test:
            stage = 'test'

        for x, y in tqdm(data_loader):

            x = x.to(DEVICE)
            y = y.to(DEVICE)

            y_prim = model.forward(x)
            loss = torch.mean(-y * torch.log(y_prim + 1e-8))

            if data_loader == dataloader_train:
                loss.backward()
                optimizer.step()
                optimizer.zero_grad()

            np_y_prim = y_prim.cpu().data.numpy()
            np_y = y.cpu().data.numpy()
            x = x.cpu()

            idx_y = np.argmax(np_y, axis=1)
            idx_y_prim = np.argmax(np_y_prim, axis=1)

            acc = np.average((idx_y == idx_y_prim) * 1.0)

            metrics_epoch[f'{stage}_acc'].append(acc)
            metrics_epoch[f'{stage}_loss'].append(loss.cpu().item())

        metrics_strs = []
        for key in metrics_epoch.keys():
            if stage in key:
                value = np.mean(metrics_epoch[key])
                metrics[key].append(value)
                metrics_strs.append(f'{key}: {round(value, 2)}')

        print(f'epoch: {epoch} {" ".join(metrics_strs)}')

    plt.clf()
    plt.subplot(121)  # row col idx
    plts = []
    c = 0
    for key, value in metrics.items():
        plts += plt.plot(value, f'C{c}', label=key)
        ax = plt.twinx()
        c += 1

    plt.legend(plts, [it.get_label() for it in plts])

    for i, j in enumerate([4, 5, 6, 10, 11, 12, 16, 17, 18]):
        plt.subplot(3, 6, j)
        color = 'green' if idx_y[i] == idx_y_prim[i] else 'red'
        plt.title(f"pred: {dataset_full.labels[idx_y_prim[i]]}\n real: {dataset_full.labels[idx_y[i]]}", color=color)
        plt.imshow(x[i].permute(1, 2, 0))

    plt.tight_layout(pad=0.5)
    plt.draw()
    plt.pause(0.1)
